#include <linux/module.h>
#include <linux/fs.h>
#include <asm/uaccess.h>

static char *hello = "hello world\n";

struct idata {
	size_t count;
};

static int driver_open( struct inode *geraetedatei, struct file *instance )
{
	struct idata *instance_data;

	instance->private_data = kmalloc(sizeof(struct idata), GFP_USER);
	if( instance->private_data==NULL ) {
		printk("driver_open: out of memory\n");
		return -EIO;
	}
	instance_data = (struct idata *)instance->private_data;
	instance_data->count = strlen(hello) + 1;
	printk("driver_open (count=%ld)\n",instance_data->count);
	return 0;
}

static int driver_release( struct inode *geraetedatei, struct file *instance )
{
	printk("driver_release\n");
	if( instance->private_data )
		kfree( instance->private_data );
	return 0;
}

static ssize_t driver_read(struct file *instance,char __user *buf,size_t count,
	loff_t *offset )
{
	struct idata *instance_data;
	int not_copied=0;
	int to_copy;

	instance_data = (struct idata *)instance->private_data;
	if( instance_data )
		to_copy = min( instance_data->count, count );
	else
		return -EFAULT;
	if( to_copy ) {
		not_copied = copy_to_user( buf, hello, to_copy );
	}
	instance_data->count -= (to_copy-not_copied);
	printk("instance_data->count %ld\n", instance_data->count );
	return to_copy-not_copied;
}

static struct file_operations fops = {
	.owner = THIS_MODULE,
	.open  = driver_open,
	.release = driver_release,
	.read = driver_read,
};

static int __init driver_init(void)
{
	printk("init_module called\n");
	if( register_chrdev( 240, "openclose", &fops )==0 ) {
		return 0;
	}
	printk("loading failed\n");
	return -EIO;
}

static void __exit driver_exit(void)
{
	unregister_chrdev( 240, "openclose" );
	printk("cleanup_module called\n");
}

module_init( driver_init );
module_exit( driver_exit );
MODULE_LICENSE("GPL");

