#include <linux/module.h>
#include <linux/fs.h>
#include <asm/uaccess.h>

/**********************************************************/
/* ACHTUNG: Diese Loesung beruecksichtigt nicht, dass der */
/* kritische Abchnitt geschuetzt werden muss!             */
/**********************************************************/

static char internal_buffer[100];
static int pos = 0;
static wait_queue_head_t wq;

static ssize_t driver_read(struct file *instanz, char __user *buf,
	size_t count, loff_t *offset )
{
	int tocopy=0, not_copied=0;

	while( pos==0 ) {
		printk("going to sleep ...\n");
		if( wait_event_interruptible(wq,(pos>0)) )
			return -ERESTART;
	}
	printk("woke up ...\n");
	tocopy=min(pos,(int)count);
	not_copied=copy_to_user(buf,internal_buffer+pos-tocopy,tocopy);
	pos = pos-tocopy;
	return tocopy-not_copied;
}

static ssize_t driver_write(struct file *instanz,const char __user *buf,
	size_t count, loff_t *offset )
{
	int maxleft, tocopy, not_copied;

	printk("driver_write...\n");
	maxleft = sizeof(internal_buffer)-pos;
	tocopy  = min( maxleft, (int)count );
	not_copied = copy_from_user( internal_buffer+pos, buf, tocopy );
	pos += tocopy-not_copied;
	wake_up( &wq );
	return tocopy-not_copied;
}

static struct file_operations fops = {
	.owner = THIS_MODULE,
	.read = driver_read,
	.write = driver_write,
};

static int __init driver_init(void)
{
	printk("driver_init called\n");
	if( register_chrdev( 240, "buf", &fops )==0 ) {
		init_waitqueue_head( &wq );
		return 0;
	}
	printk("loading failed\n");
	return -EIO;
}

static void __exit driver_exit(void)
{
	unregister_chrdev( 240, "buf" );
	printk("driver_exit called\n");
}

module_init( driver_init );
module_exit( driver_exit );
MODULE_LICENSE("GPL");
