#include <linux/module.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <asm/uaccess.h>

static char internal_buffer[100];
static int pos = 0;
static wait_queue_head_t wq;
static wait_queue_head_t wqw;
DECLARE_MUTEX( s );

#define maxleft  (sizeof(internal_buffer)-pos)

static int driver_open( struct inode *geraetedatei, struct file *instanz )
{
	printk("driver_open\n");
	return 0;
}

static int driver_release( struct inode *geraetedatei, struct file *instanz )
{
	printk("driver_release\n");
	return 0;
}

static ssize_t driver_read(struct file *instanz, char __user *buf,
	size_t count, loff_t *offset )
{
	int tocopy=0, not_copied=0;

	down_interruptible( &s );
	while( pos==0 ) {
		up( &s );
		if( instanz->f_flags & O_NONBLOCK )
			return -EAGAIN;
		printk("going to sleep ...\n");
		if( wait_event_interruptible(wq,(pos>0)) )
			return -ERESTART;
		down_interruptible( &s );
	}
	printk("woke up ...\n");
	tocopy=min(pos,(int)count);
	not_copied=copy_to_user(buf,internal_buffer+pos-tocopy,tocopy);
	pos = pos-tocopy;
	wake_up( &wqw );
	up( &s );
	return tocopy-not_copied;
}

static ssize_t driver_write(struct file *instanz,const char __user *buf,
	size_t count, loff_t *offset )
{
	int tocopy, not_copied;

	printk("driver_write...\n");
	down_interruptible( &s );
	while( maxleft<=0 ) {
		printk("setting writer to sleep ...(%ld, %d)\n", maxleft, pos);
		up( &s );
		if( wait_event_interruptible(wqw,(maxleft>0)) )
			return -ERESTART;
		down_interruptible( &s );
	}
	printk("writer woke up ...\n");
	if( maxleft > (int) count )
		tocopy = count;
	else
		tocopy = maxleft;
	not_copied = copy_from_user( internal_buffer+pos, buf, tocopy );
	pos += tocopy-not_copied;
	up( &s );
	wake_up( &wq );
	return tocopy-not_copied;
}

static struct file_operations fops = {
	.owner = THIS_MODULE,
	.open  = driver_open,
	.release = driver_release,
	.read = driver_read,
	.write = driver_write,
};

static int __init driver_init(void)
{
	printk("driver_init called\n");
	if( register_chrdev( 240, "buf", &fops )==0 ) {
		init_waitqueue_head( &wq );
		init_waitqueue_head( &wqw );
		return 0;
	}
	printk("loading failed\n");
	return -EIO;
}

static void __exit driver_exit(void)
{
	unregister_chrdev( 240, "buf" );
	printk("driver_exit called\n");
}

module_init( driver_init );
module_exit( driver_exit );
MODULE_LICENSE("GPL");
